#!/usr/bin/perl -wT

use strict;

my (%functions, %escapes, $readsofar);
my ($mtType, $mtHeaderSize, $mtVersion, $mtSize, $mtNoObjects,
    $mtMaxRecord, $mtNoParameters);

sub loadfunctions
{
    my $handle;
    if (open($handle, "<", "metafunc.txt"))
    {
	while (<$handle>)
	{
	    if (/^(.*?)\s+0x(\w\w\w\w)/)
	    {
		$functions{hex($2)} = $1;
	    }
	}
	close $handle;
    }
}

sub loadescapes
{
    my $handle;
    if (open($handle, "<", "escapes.txt"))
    {
	while (<$handle>)
	{
	    if (/^(\d+) (\w+)/)
	    {
		$escapes{$1} = $2;
	    }
	}
	close $handle;
    }
}

sub printheader
{
    my $filehandle = shift;
    my ($retval, $data, $readin);
    $retval = 0;
    $readin = read($filehandle, $data, 18);
    if ($readin == 18 && unpack("V", $data) == 0x9AC6CDD7)
    {
	print "Skipping placeable metafile header.\n";
	$readin = read($filehandle, $data, 4);
	if ($readin == 4)
	{
	    $readin = read($filehandle, $data, 18);
	}
    }
    if ($readin == 18)
    {
	($mtType, $mtHeaderSize, $mtVersion, $mtSize, $mtNoObjects,
	$mtMaxRecord, $mtNoParameters) = unpack("vvvVvVv", $data);
	printf "mtType = 0x%04X (should be 0x0001)\n", $mtType;
	printf "mtHeaderSize = 0x%04X (should be 0x0009)\n", $mtHeaderSize;
	printf "mtVersion = 0x%04X (should be 0x0300)\n", $mtVersion;
	printf "mtSize = 0x%08X (total words including this header)\n",
	    $mtSize;
	printf "mtNoObjects = 0x%04X (maximum objects at a time)\n",
	    $mtNoObjects;
	printf "mtMaxRecord = 0x%08X (maximum record size in words)\n",
	    $mtMaxRecord;
	printf "mtNoParameters = 0x%04X (not used)\n", $mtNoParameters;
	$readsofar = 18;
	$retval = 1;
    }
    else
    {
	print "short read\n"
    }
    return $retval;
}

sub printrecord
{
    my $filehandle = shift;
    my ($retval, $data, $readin, $recordsize, $toread);
    my ($function, $functionstr, $word, $i, $clipped);
    $retval = 0;
    if ($readsofar > $mtSize * 2)
    {
	print "!!! WMF already finished, following record should not be"
    	    ." here !!!\n";
    }
    if (($readin = read($filehandle, $data, 4)) == 4)
    {
	$recordsize = unpack("V", $data);
	if ($recordsize > $mtMaxRecord)
	{
	    print "!!! Following record is larger than "
		."maximum size specified in header !!!\n";
	}
	printf "0x%08X ", $recordsize;
	$toread = $recordsize * 2 - 4;
	$clipped = 0;
	if ($toread > 100)
	{
	    $toread = 10;
	    $clipped = 1;
	}
	if (read($filehandle, $data, $toread) == $toread)
	{
	    $readsofar += $toread + 4;
	    $function = unpack("v", $data);
	    $functionstr = "?";
	    if (exists($functions{$function}))
	    {
		$functionstr = $functions{$function};
	    }
	    printf "0x%04X (%s) ", $function, $functionstr;
	    $i = 2;
	    while ($i < $toread)
	    {
		$word = unpack("x$i v", $data);
		printf "0x%04X ", $word;
		if ($i == 2 && $function == 0x0626) # escape
		{
		    print "(".$escapes{$word}.") ";
		}
		$i += 2;
	    }
	    $retval = 1;
	}
	else
	{
	    print "short read";
	}
	if ($clipped)
	{
	    my $toskip;
	    print "...";
	    $toskip = $recordsize * 2 - 4 - $toread;
	    if (seek($filehandle, $toskip, 1) != 1)
	    {
		$retval = 0;
	    }
	    $readsofar += $toskip;
	}
	print "\n";
    }
    else
    {
	if ($readin != 0)
	{
	    print "short read\n";
	}
    }
    return $retval;
}

sub dumpwmf
{
    my $filename = shift;
    my $filehandle;
    print "Processing file: $filename\n";
    if (open($filehandle, "<", $filename))
    {
	print "HEADER\n";
	if (printheader($filehandle))
	{
	    print "RECORDS (length, function code, parameters)\n";
	    while (printrecord($filehandle))
	    {
	    };
	}
	close $filehandle;
    }
    else
    {
	print "Open error.\n";
    }
    print "\n";
}

my $filename;

if ($#ARGV < 0)
{
    print "Usage: $0 <filename>\n";
}
else
{
    loadfunctions();
    loadescapes();
    for $filename (@ARGV)
    {
	dumpwmf($filename);
    }
}
